/* filemode.c -- make a string describing file modes
   Copyright (C) 1985, 1990, 1993 Free Software Foundation, Inc.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "swuser_config.h"
#include <sys/types.h>
#include <sys/stat.h>

#if !S_IRUSR
# if S_IREAD
#  define S_IRUSR S_IREAD
# else
#  define S_IRUSR 00400
# endif
#endif

#if !S_IWUSR
# if S_IWRITE
#  define S_IWUSR S_IWRITE
# else
#  define S_IWUSR 00200
# endif
#endif

#if !S_IXUSR
# if S_IEXEC
#  define S_IXUSR S_IEXEC
# else
#  define S_IXUSR 00100
# endif
#endif

#ifdef STAT_MACROS_BROKEN
#undef S_ISBLK
#undef S_ISCHR
#undef S_ISDIR
#undef S_ISFIFO
#undef S_ISLNK
#undef S_ISMPB
#undef S_ISMPC
#undef S_ISNWK
#undef S_ISREG
#undef S_ISSOCK
#endif /* STAT_MACROS_BROKEN.  */

#if !defined(S_ISBLK) && defined(S_IFBLK)
#define S_ISBLK(m) (((m) & S_IFMT) == S_IFBLK)
#endif
#if !defined(S_ISCHR) && defined(S_IFCHR)
#define S_ISCHR(m) (((m) & S_IFMT) == S_IFCHR)
#endif
#if !defined(S_ISDIR) && defined(S_IFDIR)
#define S_ISDIR(m) (((m) & S_IFMT) == S_IFDIR)
#endif
#if !defined(S_ISREG) && defined(S_IFREG)
#define S_ISREG(m) (((m) & S_IFMT) == S_IFREG)
#endif
#if !defined(S_ISFIFO) && defined(S_IFIFO)
#define S_ISFIFO(m) (((m) & S_IFMT) == S_IFIFO)
#endif
#if !defined(S_ISLNK) && defined(S_IFLNK)
#define S_ISLNK(m) (((m) & S_IFMT) == S_IFLNK)
#endif
#if !defined(S_ISSOCK) && defined(S_IFSOCK)
#define S_ISSOCK(m) (((m) & S_IFMT) == S_IFSOCK)
#endif
#if !defined(S_ISMPB) && defined(S_IFMPB) /* V7 */
#define S_ISMPB(m) (((m) & S_IFMT) == S_IFMPB)
#define S_ISMPC(m) (((m) & S_IFMT) == S_IFMPC)
#endif
#if !defined(S_ISNWK) && defined(S_IFNWK) /* HP/UX */
#define S_ISNWK(m) (((m) & S_IFMT) == S_IFNWK)
#endif


void mode_string ( unsigned short mode, char * str);
static char ftypelet (long bits);
static void rwx(unsigned short bits, char *chars);
static void setst (unsigned short bits, char *chars);

/* filemodestring - fill in string STR with an ls-style ASCII
   representation of the st_mode field of file stats block STATP.
   10 characters are stored in STR; no terminating null is added.
   The characters stored in STR are:

   0	File type.  'd' for directory, 'c' for character
	special, 'b' for block special, 'm' for multiplex,
	'l' for symbolic link, 's' for socket, 'p' for fifo,
	'-' for regular, '?' for any other file type

   1	'r' if the owner may read, '-' otherwise.

   2	'w' if the owner may write, '-' otherwise.

   3	'x' if the owner may execute, 's' if the file is
	set-user-id, '-' otherwise.
	'S' if the file is set-user-id, but the execute
	bit isn't set.

   4	'r' if group members may read, '-' otherwise.

   5	'w' if group members may write, '-' otherwise.

   6	'x' if group members may execute, 's' if the file is
	set-group-id, '-' otherwise.
	'S' if it is set-group-id but not executable.

   7	'r' if any user may read, '-' otherwise.

   8	'w' if any user may write, '-' otherwise.

   9	'x' if any user may execute, 't' if the file is "sticky"
	(will be retained in swap space after execution), '-'
	otherwise.
	'T' if the file is sticky but not executable.  */

static
mode_t
set_perms(char * fps, int set, int * p_error)
{
	mode_t ret_mode;
	char * s;
	char r_b;
	char w_b;
	char x_b;	
	unsigned int m_r;
	unsigned int m_w;
	unsigned int m_x;
	unsigned int m_s;
	unsigned int m_t;

	s = fps;
	/* set == 0   is user */
	/* set == 1   is group */
	/* set == 2   is other */
	
	if (set == 0) {
		m_r = S_IRUSR;
		m_w = S_IWUSR;
		m_x = S_IXUSR;
		m_s = S_ISUID;
		m_t = 0;
	} else if (set == 1) {
		m_r = S_IRGRP;
		m_w = S_IWGRP;
		m_x = S_IXGRP;
		m_s = S_ISGID;
		m_t = 0;
	} else if (set == 2) {
		m_r = S_IROTH;
		m_w = S_IWOTH;
		m_x = S_IXOTH;
		m_s = 0;
		m_t = S_ISVTX;
	}
	
	*p_error = 0;
	if (*s == '\0') { *p_error = 1; return 0; }
	r_b = *s;
	
	s++; if (*s == '\0') { *p_error = 1; return 0; }
	w_b = *s;
	
	s++; if (*s == '\0') { *p_error = 1; return 0; }
	x_b = *s;

	ret_mode = 0;

	if (r_b == 'r') {
		ret_mode |= m_r;
	}

	if (w_b == 'w') {
		 ret_mode |= m_w;
	}

	if (x_b == 'x') {
		ret_mode |= m_x;
	}

	if (x_b == 's') {
		ret_mode |= m_x;
		ret_mode |= m_s;
	}

	if (x_b == 't') {
		ret_mode |= m_x;
		 ret_mode |= m_t;
	}
	return ret_mode;
}

void
swlib_filemodestring (mode_t  mode, char * str)
{
  mode_string ((unsigned short)mode, str);
}

void
filemodestring (struct stat * statp, char *str)
{
  mode_string (statp->st_mode, str);
}

/* Like filemodestring, but only the relevant part of the `struct stat'
   is given as an argument.  */

void
mode_string (unsigned short mode, char * str)
{
  str[0] = ftypelet ((long) mode);
  rwx ((mode & 0700) << 0, &str[1]);
  rwx ((mode & 0070) << 3, &str[4]);
  rwx ((mode & 0007) << 6, &str[7]);
  setst (mode, str);
}

/* Return a character indicating the type of file described by
   file mode BITS:
   'd' for directories
   'b' for block special files
   'c' for character special files
   'm' for multiplexor files
   'l' for symbolic links
   's' for sockets
   'p' for fifos
   '-' for regular files
   '?' for any other file type.  */

static char
ftypelet (bits)
     long bits;
{
#ifdef S_ISBLK
  if (S_ISBLK (bits))
    return 'b';
#endif
  if (S_ISCHR (bits))
    return 'c';
  if (S_ISDIR (bits))
    return 'd';
  if (S_ISREG (bits))
    return '-';
#ifdef S_ISFIFO
  if (S_ISFIFO (bits))
    return 'p';
#endif
#ifdef S_ISLNK
  if (S_ISLNK (bits))
    return 'l';
#endif
#ifdef S_ISSOCK
  if (S_ISSOCK (bits))
    return 's';
#endif
#ifdef S_ISMPC
  if (S_ISMPC (bits))
    return 'm';
#endif
#ifdef S_ISNWK
  if (S_ISNWK (bits))
    return 'n';
#endif

#ifdef S_ISOFD
  /* Cray migrated dmf file.  */
  if (S_ISOFD (bits))
    return 'M';
#endif
#ifdef S_ISOFL
  /* Cray migrated dmf file.  */
  if (S_ISOFL (bits))
    return 'M';
#endif
  return '?';
}

/* Look at read, write, and execute bits in BITS and set
   flags in CHARS accordingly.  */

static void
rwx (unsigned short bits, char * chars)
{
  chars[0] = (bits & S_IRUSR) ? 'r' : '-';
  chars[1] = (bits & S_IWUSR) ? 'w' : '-';
  chars[2] = (bits & S_IXUSR) ? 'x' : '-';
}

/* Set the 's' and 't' flags in file attributes string CHARS,
   according to the file mode BITS.  */

static void
setst (unsigned short bits, char * chars)
{
#ifdef S_ISUID
  if (bits & S_ISUID)
    {
      if (chars[3] != 'x')
	/* Set-uid, but not executable by owner.  */
	chars[3] = 'S';
      else
	chars[3] = 's';
    }
#endif
#ifdef S_ISGID
  if (bits & S_ISGID)
    {
      if (chars[6] != 'x')
	/* Set-gid, but not executable by group.  */
	chars[6] = 'S';
      else
	chars[6] = 's';
    }
#endif
#ifdef S_ISVTX
  if (bits & S_ISVTX)
    {
      if (chars[9] != 'x')
	/* Sticky, but not executable by others.  */
	chars[9] = 'T';
      else
	chars[9] = 't';
    }
#endif
}

mode_t
swlib_filestring_to_mode(char * str, int do_include_type)
{
	mode_t ret;
	mode_t mode_ret;
	int error;
	char * s;
	int type;

	mode_ret = (mode_t)0;
	s = str;

	mode_ret = (mode_t)0;
	type = *s;	
	if (
		type == 'd' ||
		type == '-' ||
		type == 'l' ||
		type == 's' 
	) {
		if (do_include_type) {
			/* FIXME, set type */
		}
		;
	} else {
		/* error */
		return 0;
	}
	
	s++; if (*s == '\0') return 0;
	ret = set_perms(s, 0, &error);  /* user */
	if (error) return 0;
	mode_ret |= ret;

	s++; if (*s == '\0') return 0;
	s++; if (*s == '\0') return 0;
	s++; if (*s == '\0') return 0;
	ret = set_perms(s, 1, &error);  /* group */
	if (error) return 0;
	mode_ret |= ret;
	
	s++; if (*s == '\0') return 0;
	s++; if (*s == '\0') return 0;
	s++; if (*s == '\0') return 0;
	ret = set_perms(s, 2, &error);  /* other */
	if (error) return 0;
	mode_ret |= ret;

	return mode_ret;
}
